/*
  Copyright (C) 2009 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*! Last updated: 16-03-05 : RB >> made light ID assignment automatic */
#include "Light.h"

//----------------------------------------------------------------------------------------------------------------------
/// @file Light.cpp
/// @brief implementation files for Light class
//----------------------------------------------------------------------------------------------------------------------
namespace ngl
{

/// all possible light ID's. Wheh lights are enabled, they simply take
/// the last first available light ID. When they get disabled, the light
/// ID gets written back into this array.
///
unsigned char Light::s_freeLights[8] = {0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08};
unsigned int Light::s_numEnabledLights=0;

//----------------------------------------------------------------------------------------------------------------------
Light::Light(
             const Vector& _pos,
             const Colour& _col,
             const LIGHTMODES _lightMode
            )
{
	// zero now means un-assigned
	m_lightNo=0x00;
	m_pos.set(_pos);
	m_colour.set(_col);
	m_specColour.set(_col);
	m_lightMode = _lightMode;
  m_constantAtten=1.0;
  m_linearAtten=0.0;
  m_quadraticAtten=0.0;
}

//----------------------------------------------------------------------------------------------------------------------
Light::Light(
             const Vector& _pos,
             const Colour& _col,
             const Colour& _specColour,
             const LIGHTMODES _lightMode)
{
	// zero now means un-assigned
	m_lightNo=0x00;
	m_pos.set(_pos);
	m_colour.set(_col);
	m_specColour.set(_specColour);
	m_lightMode=_lightMode;
  m_constantAtten=1.0;
  m_linearAtten=0.0;
  m_quadraticAtten=0.0;

}

//----------------------------------------------------------------------------------------------------------------------
void Light::show()
{
	glPushMatrix();
		m_pos.translate();
		m_colour.use();
		VBOPrimitives *prim=VBOPrimitives::instance();
		glScalef(0.2,0.2,0.2);
		prim->draw("cube");
	glPopMatrix();
}

//----------------------------------------------------------------------------------------------------------------------
bool Light::enable()
{
	if (m_lightNo==0)
	{
		// if all 8 lights enabled, ignore request
		if (s_numEnabledLights==8)
		{
			std::cerr << "[WARNING] trying to enable more than 8 lights at once!" << std::endl;
			return false;
		}
		// set light num
		m_lightNo = s_freeLights[s_numEnabledLights++];
	}
	// set GL params
	Real zero[] = {0.0f,0.0f,0.0f,0.0f};

	GLint light=GL_LIGHT0+(m_lightNo-1);
	glLightfv(light,GL_DIFFUSE,m_colour.openGL()); // tell openGL diffuse light color
	glLightfv(light,GL_SPECULAR,m_specColour.openGL()); // tell openGl specular light color
	glLightfv(light,GL_AMBIENT,zero);
	// opengl wants a homogenous point for the light so we build one
	m_pos.m_w=(Real)m_lightMode;
	glLightf(light, GL_SPOT_CUTOFF, 180);
	glLightfv(light,GL_POSITION,m_pos.openGL());
	glLightf(light, GL_CONSTANT_ATTENUATION, m_constantAtten);
	glLightf(light, GL_LINEAR_ATTENUATION, m_linearAtten);
	glLightf(light, GL_QUADRATIC_ATTENUATION, m_quadraticAtten);

	glEnable(light);

	return true;
}

//----------------------------------------------------------------------------------------------------------------------
void Light::disable()
{
	if (m_lightNo!=0)
	{
		glDisable( getLightNum() );
		s_freeLights[--s_numEnabledLights] = m_lightNo;
		//std::cout<<"Num enabled lights"<<s_numEnabledLights<<"\n";
    m_lightNo = 0x00;
	}
}

//----------------------------------------------------------------------------------------------------------------------
void Light::setAttenuation(
                          const Real _constant,
                          const Real _linear,
                          const Real _quadratic
                          )
{
	m_constantAtten=_constant;
	m_linearAtten=_linear;
	m_quadraticAtten=_quadratic;
}



} // end ngl namespacee
