/*
  Copyright (C) 2009 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cstdlib>
#include <QDebug>
#include "ShaderManager.h"

//----------------------------------------------------------------------------------------------------------------------
/// @file ShaderManager.cpp
/// @brief implementation files for ShaderManager class
//----------------------------------------------------------------------------------------------------------------------

namespace ngl
{


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParamFromMatrix(
                                            const std::string &_shaderName,
                                            const std::string &_paramName,
                                            ngl::Matrix _p1
                                           )
{
	(*this)[_shaderName]->setUniformMatrix4fv(_paramName.c_str(),1,GL_FALSE,_p1.openGL());
}



//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParamFromVector(
                              const std::string &_shaderName,
                              const std::string &_paramName,
                              ngl::Vector _p1
                             )
{

	(*this)[_shaderName]->setUniform4fv(_paramName.c_str(),1,_p1.openGL());

}

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParam4f(
                                      const std::string &_shaderName,
                                      const std::string &_paramName,
                                      float _p1,
                                      float _p2,
                                      float _p3,
                                      float _p4

                                      )
{
	(*this)[_shaderName]->setUniform4f(_paramName.c_str(),_p1,_p2,_p3,_p4);
}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParam3f(
                                      const std::string &_shaderName,
                                      const std::string &_paramName,
                                      float _p1,
                                      float _p2,
                                      float _p3
                                    )
{
	(*this)[_shaderName]->setUniform3f(_paramName.c_str(),_p1,_p2,_p3);
}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParam2f(
                                      const std::string &_shaderName,
                                      const std::string &_paramName,
                                      float _p1,
                                      float _p2
                                    )
{
	(*this)[_shaderName]->setUniform2f(_paramName.c_str(),_p1,_p2);

}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParam1i(
                                      const std::string &_shaderName,
                                      const std::string &_paramName,
                                      int _p1
                                    )
{
	(*this)[_shaderName]->setUniform1i(_paramName.c_str(),_p1);

}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::setShaderParam1f(
                                      const std::string &_shaderName,
                                      const std::string &_paramName,
                                      float _p1
                                    )
{
	(*this)[_shaderName]->setUniform1f(_paramName.c_str(),_p1);

}

//----------------------------------------------------------------------------------------------------------------------
// this is marked for deprecation!
void ShaderManager::useShader(
                              const std::string &_name
                             )
{
 use(_name);
}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::loadShader(
                               const std::string &_shaderName,
                               const std::string &_vert,
                               const std::string &_frag,
                               const std::string &_geo,
                               const bool _exitOnError
                              )
{

  createShaderProgram(_shaderName);

  attachShader(_shaderName+"Vertex",VERTEX);
  attachShader(_shaderName+"Fragment",FRAGMENT);
  loadShaderSource(_shaderName+"Vertex",_vert);
  loadShaderSource(_shaderName+"Fragment",_frag);

  compileShader(_shaderName+"Vertex");
  compileShader(_shaderName+"Fragment");
  attachShaderToProgram(_shaderName,_shaderName+"Vertex");
  attachShaderToProgram(_shaderName,_shaderName+"Fragment");
	if( _geo !="")
	{
		attachShader(_shaderName+"Geo",GEOMETRY);
		loadShaderSource(_shaderName+"Geo",_vert);
		compileShader(_shaderName+"Geo");
		attachShaderToProgram(_shaderName,_shaderName+"Geo");
	}

  linkProgramObject(_shaderName);
}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::reset()
{
	std::cerr<<"Closing down shader manager\n";

  std::map <std::string,ShaderProgram *>::iterator pbegin=m_shaderPrograms.begin();
  std::map <std::string,ShaderProgram *>::iterator pend=m_shaderPrograms.end();
  // delete each of the shader programs first (this will clear the maps in Program)
  // but not delete the shaders
  while(pbegin != pend)
  {
    delete pbegin->second;
    ++pbegin;
  }
  // now we delete all of the shaders
  std::map <std::string,Shader *>::iterator sbegin=m_shaders.begin();;
  std::map <std::string,Shader *>::iterator send=m_shaders.end();;

  while(sbegin != send)
  {
    delete sbegin->second;
    ++sbegin;
  }

}


//----------------------------------------------------------------------------------------------------------------------
GLint ShaderManager::getAttribLocation(
                                        const std::string &_shaderName,
                                        const std::string &_paramName
                                       )
{

  GLint attrib=0;

  // get an iterator to the shaders
  std::map <std::string, ShaderProgram * >::const_iterator shader=m_shaderPrograms.find(_shaderName);
  // make sure we have a valid shader
	if(shader!=m_shaderPrograms.end())
  {
    // grab the pointer to the shader and call compile
    attrib=glGetAttribLocation((long) shader->second->getID(),_paramName.c_str());
  }
	else
	{
		std::cerr <<"Warning trying to get attrib "<<_paramName<<" from "<<_shaderName<<" but not found\n";
	}

  return attrib;

}



//----------------------------------------------------------------------------------------------------------------------
ShaderManager::ShaderManager()
{
//std::cout<<"Shader Manager ctor\n";
 m_debugState=true;
 m_numShaders=0;
 m_nullProgram = new ShaderProgram("NULL");
}
//----------------------------------------------------------------------------------------------------------------------

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::attachShader(
                                  std::string _name,
                                  SHADERTYPE _type
                                )
{
  m_shaders[_name]= new Shader(_name,_type);
}

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::compileShader(
																	 std::string _name
																	)
{
  // get an iterator to the shaders
  std::map <std::string, Shader * >::const_iterator shader=m_shaders.find(_name);
  // make sure we have a valid shader
	if(shader!=m_shaders.end())
  {
    // grab the pointer to the shader and call compile
    shader->second->compile();
  }
  else {std::cerr<<"Warning shader not know in compile "<<_name.c_str();}

}
//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::createShaderProgram(
																				 std::string _name
																			  )
{
 std::cerr<<"creating empty ShaderProgram "<<_name.c_str()<<"\n";
 m_shaderPrograms[_name]= new ShaderProgram(_name);
}
//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::attachShaderToProgram(
																					 std::string _program,
																					 std::string _shader
																					)

{

  // get an iterator to the shader and program
  std::map <std::string, Shader * >::const_iterator shader=m_shaders.find(_shader);
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_program);

  // make sure we have a valid shader and program
	if(shader!=m_shaders.end() && program !=m_shaderPrograms.end())
  {
    // now attach the shader to the program
    program->second->attatchShader(shader->second);
    // now increment the shader ref count so we know if how many references
    shader->second->incrementRefCount();

    if (m_debugState == true)
    {
      std::cerr<<_shader.c_str()<<" attached to program "<<_program.c_str()<<"\n";
    }
  }
  else {std::cerr<<"Warning cant attach "<<_shader.c_str() <<" to "<<_program.c_str()<<"\n";}
}

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::loadShaderSource(
																			std::string _shaderName,
																			std::string _sourceFile
																		 )
{
  std::map <std::string, Shader * >::const_iterator shader=m_shaders.find(_shaderName);
  // make sure we have a valid shader and program
	if(shader!=m_shaders.end() )
  {
    shader->second->load(_sourceFile);
  }
  else {std::cerr<<"Warning shader not know in loadShaderSource "<<_shaderName.c_str();}

}

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::linkProgramObject(
																			 std::string _name
																			)
{

  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_name);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
    std::cerr<<"Linking "<<_name.c_str()<<"\n";
    program->second->link();
  }
  else {std::cerr<<"Warning Program not known in link "<<_name.c_str();}

}

//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::use(
												 std::string _name
												)
{
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_name);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
    //std::cerr<<"Shader manager Use\n";
    program->second->use();
  }
  else
  {
    std::cerr<<"Warning Program not know in use "<<_name.c_str();
    glUseProgram(0);
  }

}


//----------------------------------------------------------------------------------------------------------------------
GLuint ShaderManager::getProgramID(
																		std::string _name
																	 )
{
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_name);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
    return program->second->getID();
  }
  else
  {
    std::cerr<<"Warning Program not know in use "<<_name.c_str();
    return -1;
  }
}


//----------------------------------------------------------------------------------------------------------------------
void ShaderManager::bindAttribute(
																	 std::string _programName,
																	 GLuint _index,
																	 std::string _attribName
																  )
{
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_programName);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
    program->second->bindAttrib(_index,_attribName);
  }
  else {std::cerr<<"Warning Program not know in bindAttrib "<<_programName.c_str();}
}


//----------------------------------------------------------------------------------------------------------------------
void ceckGLError(
                  const std::string  &_file,
                  const int _line
                )
{
 int err_num = glGetError();

	while (err_num != GL_NO_ERROR)
  {
    const char* str = (const char*)gluErrorString(err_num);
    QString message =QString("GL: %1  line: %2, file: %3").arg( str).arg( _line).arg( _file.c_str());
    qDebug()<<message;
    err_num = glGetError();
  }
}


//----------------------------------------------------------------------------------------------------------------------
ShaderProgram * ShaderManager::operator[](
																					 const std::string &_programName
																					)
{
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_programName);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
		return  program->second;
  }
  else
  {
    std::cerr<<"Warning Program not know in [] "<<_programName.c_str();
    std::cerr<<"returning a null program and hoping for the best\n";
    return m_nullProgram;
  }
}


//----------------------------------------------------------------------------------------------------------------------
ShaderProgram * ShaderManager::operator[](
																					const char *_programName
																					)
{
  std::map <std::string, ShaderProgram * >::const_iterator program=m_shaderPrograms.find(_programName);
  // make sure we have a valid  program
	if(program!=m_shaderPrograms.end() )
  {
    return  program->second;
  }
  else
  {
    std::cerr<<"Warning Program not know in [] "<<_programName;
    std::cerr<<"returning a null program and hoping for the best\n";
    return m_nullProgram;
  }
}


void ShaderManager::useNullProgram()
{
  m_nullProgram->use();

}


} // end ngl namespace



//----------------------------------------------------------------------------------------------------------------------

